/**
 *  @file       log/frame.hpp
 *  @author     Taka
 *  @date       23oct97
 *  
 *  @copyright  (c) T. Muraoka 1997 (all rights reserved).
 *              Unauthorized use of this code is prohibited.
 *
 *  @brief      Defines the CMessageLogFrame class.
 * 
 *  @licence    This software is provided 'as-is', without any express 
 *              or implied warranty.
 *
 *              In no event will the author be held liable for any damages 
 *              arising from the use of this software.
 *  
 *              Permission is granted to anyone to use this software 
 *              for any purpose, including commercial applications, 
 *              and to alter it and redistribute it freely, subject to 
 *              the following restrictions: 
 *
 *              - The origin of this software must not be misrepresented; 
 *                you must not claim that you wrote the original software. 
 *                If you use this software in a product, an acknowledgment 
 *                in the product documentation is requested but not required. 
 *
 *              - Altered source versions must be plainly marked as such, 
 *                and must not be misrepresented as being the original software. 
 *                Altered source is encouraged to be submitted back to 
 *                the original author so it can be shared with the community. 
 *                Please share your changes. 
 *
 *              - This notice may not be removed or altered from any 
 *                source distribution. 
 */

#ifndef LOG_FRAME_HPP
#define LOG_FRAME_HPP

#include "common.hpp"

namespace debug
{
namespace log
{

/* --- CLASS DEFINITION: CMessageLogFrame ----------------------------- */

/// Manages indentation of log messages.
/** @class CMessageLogFrame log/frame.hpp "log/frame.hpp"
 *
 *  When an instance of this class is created, messages will be tabbed out 
 *  one extra tabstop for the specified CMessageLog. The destructor will 
 *  un-indent back, thus ensuring that this always gets done.
 *
 *  @see BEGIN_MESSAGE_LOG_FRAME(), END_MESSAGE_LOG_FRAME()
 */

class CMessageLogFrame
{

// constructors:
public:
    CMessageLogFrame( CMessageLog& msgLog ) : mMsgLog(msgLog) { mMsgLog.beginFrame() ; }
// destructor:
    ~CMessageLogFrame() { mMsgLog.endFrame() ; }

// data members:
private:
    CMessageLog& mMsgLog ;

} ;

} // End namespace log
} // End namespace debug

/* --- MACROS --------------------------------------------------------- */

/// @name Message log frame macros
/** These macros are provided to help set up nested log message frames
 *  which can be useful when tracing deeply nested or recursive function 
 *  calls.
 *  For example:
 *
 *  @code
 *  void foo1()
 *  {
 *      theMessageLog() << "Entered foo1()" << endl ;
 *      BEGIN_MESSAGE_LOG_FRAME( theMessageLog() )
 *      {
 *          theMessageLog() << "foo1() processing..." << endl ;
 *          foo2() ;
 *      }
 *      END_MESSAGE_LOG_FRAME( theMessageLog() ) 
 *      theMessageLog() << "Leaving foo1()" << endl ;
 *  }
 *  
 *  void foo2()
 *  {
 *      theMessageLog() << "Entered foo2()" << endl ;
 *      BEGIN_MESSAGE_LOG_FRAME( theMessageLog() ) 
 *      {
 *          theMessageLog() << "foo2() processing..." << endl ;
 *      }
 *      END_MESSAGE_LOG_FRAME( theMessageLog() ) 
 *      theMessageLog() << "Leaving foo2()" << endl ;
 *  }
 *  @endcode
 *
 *  This will produce the following output:
 *
 *  @code
 *  Entered foo1()
 *    foo1() processing...
 *    Entered foo2()
 *      foo2() processing...
 *    Leaving foo2()
 *  Leaving foo1()
 *  @endcode
 */
//@{

#ifdef _LOG

    /** Begins a new message log frame. */
    #define BEGIN_MESSAGE_LOG_FRAME( msgLog ) \
    { \
        CMessageLogFrame x_msgLogFrame( msgLog ) ;

    /** Ends the current message log frame. */
    #define END_MESSAGE_LOG_FRAME( msgLog ) \
    }

#else 

    #define BEGIN_MESSAGE_LOG_FRAME( msgLog ) { }
    #define END_MESSAGE_LOG_FRAME( msgLog ) { }

#endif // _LOG

//@}

/* -------------------------------------------------------------------- */

#endif // LOG_FRAME_HPP
