/** @if INTERNAL
 *
 *  @file       src/log.cpp
 *  @author     Taka
 *  @date       23nov97
 *  
 *  @copyright  (c) T. Muraoka 1997 (all rights reserved).
 *              Unauthorized use of this code is prohibited.
 *  
 *  @brief      Implements the CMessageLog class.
 * 
 *  @licence    This software is provided 'as-is', without any express 
 *              or implied warranty.
 *
 *              In no event will the author be held liable for any damages 
 *              arising from the use of this software.
 *  
 *              Permission is granted to anyone to use this software 
 *              for any purpose, including commercial applications, 
 *              and to alter it and redistribute it freely, subject to 
 *              the following restrictions: 
 *
 *              - The origin of this software must not be misrepresented; 
 *                you must not claim that you wrote the original software. 
 *                If you use this software in a product, an acknowledgment 
 *                in the product documentation is requested but not required. 
 *
 *              - Altered source versions must be plainly marked as such, 
 *                and must not be misrepresented as being the original software. 
 *                Altered source is encouraged to be submitted back to 
 *                the original author so it can be shared with the community. 
 *                Please share your changes. 
 *
 *              - This notice may not be removed or altered from any 
 *                source distribution. 
 *  
 *  @endif
 */

#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <cctype>
#include <iostream>
#include <cassert>

#include "log.h"
#include "msgGroup.h"

using namespace std ;

namespace debug
{
namespace log
{

/* --- CLASS DATA ----------------------------------------------------- */

#ifdef _DEBUG
// NOTE: This flag allows us to output fixed-string line prefix information,
// which makes testing easier.
bool    CMessageLog::mFixedLinePrefix = false ;
#endif // _DEBUG

#ifndef _MAX_PATH
#define    _MAX_PATH               260
#define    _MAX_DRIVE              3 
#define    _MAX_DIR                256
#define    _MAX_FNAME              256
#define    _MAX_EXT                256
#endif

/* --- CONSTRUCTOR ---------------------------------------------------- */

/// @constructors
//@{

/** @if INTERNAL
 *  The CMessageLog class is a wrapper around an ostream. CMessageLog
 *  itself is an ostream-derived class, but installs a proxy streambuf 
 *  buffer. Whenever characters are streamed into this buffer, it simply 
 *  passes them back to CMessageLog object for processing. The CMessageLog 
 *  passes these characters on to the real ostream object, but at the 
 *  same time, watches for '\n'. When it sees one, it knows that it's 
 *  time to output the line prefix information.
 *  @endif
 */

#ifdef _MSC_VER
#pragma warning (disable: 4355) // 'this' : used in base member initializer list
#endif // _MSC_VER

// NOTE: Passing mStreamBuf in to the ostream constructor is a bit naughty 
//  since it officially hasn't been constructed yet but it'll probably work...

CMessageLog::CMessageLog( ostream& os )
    : ostream( &mStreamBuf )
    , mStreamBuf( *this )
{
    // initialize data members
    mMsgLogEnabled = true ;

    // initialize data members
    mMsgGroupIDs = false ;
    mMsgNumbers = false ;
    mTimeStamps = false ;
    mDateStamps = false ;
    mFileStamps = false ;

    // initialize data members
    mMsgNo = 0 ;
    mpCurrMsgGroup = NULL ;
    mpCurrFileName = NULL ;
    mCurrLineNo = 0 ;

    // initialize data members
    mFrameDepth = 0 ; 
    mFrameTabSize = DEFAULT_FRAME_TAB_SIZE ;

    // initialize data members
    mIsNewLine = true ;
    mpRealStreamBuf = NULL ;

    // configure our proxy streambuf
    setOutputStream( os ) ;
}

//@}

/* --- DESTRUCTOR ----------------------------------------------------- */

/// @destructor
//@{

CMessageLog::~CMessageLog()
{
}

//@}

/* -------------------------------------------------------------------- */

/// @name Message frame methods
//@{

/** Increases the message indentation by one tabstop. 
 *
 *  @see endFrame(), setFrameTabSize(), BEGIN_MESSAGE_LOG_FRAME()
 */

void
CMessageLog::beginFrame()
{
    mFrameDepth ++ ;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */

/** Decreases the message indentation by one tabstop.
 *
 *  @see beginFrame(), END_MESSAGE_LOG_FRAME()
 */

void
CMessageLog::endFrame()
{
    assert( mFrameDepth > 0 /*, "Invalid frame depth: " << mFrameDepth*/ ) ;
    mFrameDepth -- ;
}

//@}

/* -------------------------------------------------------------------- */

void
CMessageLog::setOutputStream( ostream& os )
{
    mpRealStreamBuf = os.rdbuf() ; // this is where we should send output characters 
}

/* -------------------------------------------------------------------- */

int
CMessageLog::onStreamBufChar( int ch )
{
    bool    isEnabled ;
    int     rc ;

    // determine if we are currently enabled 
    isEnabled = mMsgLogEnabled && (mpCurrMsgGroup == NULL || mpCurrMsgGroup->isMsgGroupEnabled()) ;

    // check if we are at the start of a new line 
    if ( mIsNewLine )
    {
        mIsNewLine = false ;

        // output the line prefix information (for non-blank lines only)
        if ( isEnabled && ch != '\n' )
        {
            mMsgNo ++ ;
            outputLinePrefix( *((ostream *)this) ) ;
        }
    }

    // process the character 
    if ( isEnabled )
    {
        // send the character to the real streambuf for output
        rc = mpRealStreamBuf->sputc( ch ) ;
    }
    else 
    {
        // message log is disabled - just discard the character 
        rc = ch ; 
    }

    // check if we're at the end of the line 
    if ( ch == '\n' )
        mIsNewLine = true ;

    return rc ;
}

/* -------------------------------------------------------------------- */

void
CMessageLog::outputLinePrefix( ostream& os )
{
    time_t      now ;
    const char* pNow ;
    //char        driveName[ _MAX_DRIVE ] ;
    //char        dirName[ _MAX_DIR ] ;
    //char        fileName[ _MAX_FNAME  ] ;
    //char        fileExtn[ _MAX_EXT  ] ;
    char        buf[ 40 ] ;
    int         i ;

    // get the current time 
    time( &now ) ;
    pNow = ctime( &now ) ;
    assert( pNow != NULL /*, "Couldn't get the current time." */) ;

#ifdef _DEBUG
    if ( mFixedLinePrefix )
    {
        pNow = "DDD MMM DD HH:MM:SS YYYY\n" ;
        mpCurrFileName = "FILE.CPP" ;
        mCurrLineNo = 0 ;
    }
#endif // _DEBUG

    // check if message group ID's are enabled
    if ( mMsgGroupIDs ) 
    {   
        if ( mpCurrMsgGroup == NULL )
            os << "      " ;
        else 
        {
            sprintf( buf , "[%03d] " , mpCurrMsgGroup->msgGroupId() ) ;
            os << buf ;
        }
    }

    // check if message numbers are enabled
    if ( mMsgNumbers ) 
    {   
        sprintf( buf , "%06lu" , mMsgNo ) ;
        os << buf ;
        os << " " ;
    }

    // check if date stamps are enabled 
    if ( mDateStamps ) 
    {   
        if ( pNow[8] == ' ' )
            os << '0' << pNow[9] ;
        else 
            os << pNow[8] << pNow[9] ;
        os << (char)tolower(pNow[4]) << (char)tolower(pNow[5]) << (char)tolower(pNow[6]) ;
        os << pNow[22] << pNow[23] ;
        os << " " ;
    }
    
    // check if time stamps are enabled 
    if ( mTimeStamps ) 
    {   
        os << pNow[11] << pNow[12] << ":" << pNow[14] << pNow[15] << ":" << pNow[17] << pNow[18] ;
        os << " " ;
    }
    
    // check if file stamps are enabled 
    if ( mFileStamps )
    {   
        if ( mpCurrFileName == NULL )
            strcpy( buf , "         ???:   ???" ) ;
        else 
        {
            //_splitpath( (char *)mpCurrFileName , driveName , dirName , fileName , fileExtn ) ;
            //sprintf( buf , "%8s%-4s:%06lu" , fileName , fileExtn , mCurrLineNo ) ;
            sprintf( buf , "%s" , (char *)mpCurrFileName) ;
            //strlwr( buf ) ;
        }
        os << buf ;
        os << " " ;
    }

    // terminate the message prefix information 
    if ( mMsgNumbers || mDateStamps || mTimeStamps || mFileStamps )
        os << "| " ;

    // tab out nested frames 
    for ( i=0 ; i < mFrameDepth * mFrameTabSize ; i++ )
        os << " " ;
}   

/* -------------------------------------------------------------------- */

#ifndef _DOXYGEN

void
CMessageLog::setState( CMessageGroup* pMsgGroup , const char* pFileName , size_t lineNo ) 
{ 
    // set the current state of the message log 
    mpCurrMsgGroup = pMsgGroup ; 
    mpCurrFileName = pFileName ; 
    mCurrLineNo = lineNo ; 
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */

CMessageLog&    CMessageLog::operator()( CMessageGroup* p , const char* q , size_t i ) { setState(p,q,i) ; return *this ; }

#endif // ! _DOXYGEN

/* -------------------------------------------------------------------- */

/// @name Configuration
//@{

void    CMessageLog::enableMsgGroupIDs( bool b ) { mMsgGroupIDs=b ; }
void    CMessageLog::enableMsgNumbers( bool b ) { mMsgNumbers=b ; }
void    CMessageLog::enableTimeStamps( bool b ) { mTimeStamps=b ; }
void    CMessageLog::enableDateStamps( bool b ) { mDateStamps=b ; }
void    CMessageLog::enableFileStamps( bool b ) { mFileStamps=b ; }

/** @see beginFrame(), endFrame(), BEGIN_MESSAGE_LOG_FRAME()
 */
void    CMessageLog::setFrameTabSize( int n ) { mFrameTabSize=n ; }

//@}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */

void    CMessageLog::enableMsgLog( bool b ) { mMsgLogEnabled=b ; }
bool    CMessageLog::isMsgLogEnabled() const { return mMsgLogEnabled ; }

} // End namespace log
} // End namespace debug
