/** @if INTERNAL
 *
 *  @file       src/msgGroup.cpp
 *  @author     Taka
 *  @date       25sep02
 *  
 *  @copyright  (c) T. Muraoka 1997-2002 (all rights reserved).
 *              Unauthorized use of this code is prohibited.
 *  
 *  @brief      Implements the CMessageGroup class.
 * 
 *  @licence    This software is provided 'as-is', without any express 
 *              or implied warranty.
 *
 *              In no event will the author be held liable for any damages 
 *              arising from the use of this software.
 *  
 *              Permission is granted to anyone to use this software 
 *              for any purpose, including commercial applications, 
 *              and to alter it and redistribute it freely, subject to 
 *              the following restrictions: 
 *
 *              - The origin of this software must not be misrepresented; 
 *                you must not claim that you wrote the original software. 
 *                If you use this software in a product, an acknowledgment 
 *                in the product documentation is requested but not required. 
 *
 *              - Altered source versions must be plainly marked as such, 
 *                and must not be misrepresented as being the original software. 
 *                Altered source is encouraged to be submitted back to 
 *                the original author so it can be shared with the community. 
 *                Please share your changes. 
 *
 *              - This notice may not be removed or altered from any 
 *                source distribution. 
 *  
 *  @endif
 */

#pragma warning( disable: 4786 ) // identifier was truncated to '255' characters in the debug information

#include <map>
#include <list>
#include <cassert>
#include <iostream>

#include "msgGroup.h"

using namespace std ;

namespace debug
{
namespace log
{

typedef map<int,CMessageGroup*> IntMsgGroupMap ; 
typedef map<string,CMessageGroup*> CiStringMsgGroupMap ; 
typedef list<CMessageGroup*> MsgGroupList ;

/* --- LOCAL DATA ----------------------------------------------------- */

static int 
getNextMsgGroupId()
{
    // return the next available auto-assigned message group ID
    static int nextMsgGroupId = 1 ;
    return nextMsgGroupId++ ; 
}

/* --- LOCAL DATA ----------------------------------------------------- */

static IntMsgGroupMap&      theMsgGroupIndex() { static IntMsgGroupMap index ; return index ; }
static CiStringMsgGroupMap& theMsgGroupIndexByName() { static CiStringMsgGroupMap index ; return index ; }
static MsgGroupList&        theMsgGroupList() { static MsgGroupList lst ; return lst ; }

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */

static void
registerMsgGroup( CMessageGroup* pMsgGroup )
{
    assert( pMsgGroup != NULL ) ; 

    // make sure that the message group ID is unique
    IntMsgGroupMap::const_iterator it = theMsgGroupIndex().find( pMsgGroup->msgGroupId() ) ;
    assert( it == theMsgGroupIndex().end() /*, "Duplicate message group ID: " << pMsgGroup->msgGroupId()*/ ) ;
    theMsgGroupIndex()[ pMsgGroup->msgGroupId() ] = pMsgGroup ; 
    theMsgGroupList().push_back( pMsgGroup ) ; 

    // make sure that the message group name is unique
    if ( ! pMsgGroup->msgGroupName().empty() ) 
    {
        CiStringMsgGroupMap::const_iterator it = theMsgGroupIndexByName().find( pMsgGroup->msgGroupName() ) ;
        assert( it == theMsgGroupIndexByName().end() /*, "Duplicate message group name: " << pMsgGroup->msgGroupName()*/ ) ;
        theMsgGroupIndexByName()[ pMsgGroup->msgGroupName() ] = pMsgGroup ; 
    }
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */

static void
unregisterMsgGroup( CMessageGroup* pMsgGroup )
{
    assert( pMsgGroup != NULL ) ; 

    // un-register the message group
    theMsgGroupIndex().erase( pMsgGroup->msgGroupId() ) ;
    theMsgGroupList().remove( pMsgGroup ) ; 
    if ( ! pMsgGroup->msgGroupName().empty() ) 
        theMsgGroupIndexByName().erase( pMsgGroup->msgGroupName() ) ;
}

/* --- CONSTRUCTOR ---------------------------------------------------- */

/// @constructors
//@{

CMessageGroup::CMessageGroup()
    : mMsgGroupId(getNextMsgGroupId())
    , mMsgGroupName()
    , mMsgGroupDescription()
    , mIsEnabled(true)
{
    // register the message group
    registerMsgGroup( this ) ; 
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */

CMessageGroup::CMessageGroup( const string& msgGroupName , const string& msgGroupDescription )
    : mMsgGroupId(getNextMsgGroupId())
    , mMsgGroupName(msgGroupName)
    , mMsgGroupDescription(msgGroupDescription)
    , mIsEnabled(true)
{
    // register the message group
    registerMsgGroup( this ) ; 
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */

CMessageGroup::CMessageGroup( int msgGroupId , const string& msgGroupName , const string& msgGroupDescription )
    : mMsgGroupId(msgGroupId)
    , mMsgGroupName(msgGroupName)
    , mMsgGroupDescription(msgGroupDescription)
    , mIsEnabled(true)
{
    // register the message group
    registerMsgGroup( this ) ; 
}

//@}

/* --- DESTRUCTOR ----------------------------------------------------- */

/// @destructor
//@{

CMessageGroup::~CMessageGroup()
{
    // un-register the message group
    unregisterMsgGroup( this ) ; 
}

//@}

/* -------------------------------------------------------------------- */

bool
CMessageGroup::isMsgGroupEnabled( int msgGroupId )
{
    // check if the specified message group is enabled 
    CMessageGroup* pMsgGroup = findMsgGroup( msgGroupId ) ; 
    if ( pMsgGroup == NULL )
        return false ;
    return pMsgGroup->isMsgGroupEnabled() ;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */

bool
CMessageGroup::isMsgGroupEnabled( const string& msgGroupName )
{
    // check if the specified message group is enabled 
    CMessageGroup* pMsgGroup = findMsgGroup( msgGroupName ) ; 
    if ( pMsgGroup == NULL )
        return false ;
    return pMsgGroup->isMsgGroupEnabled() ;
}

/* -------------------------------------------------------------------- */

void
CMessageGroup::enableAllMsgGroups( bool enable )
{
    // enable all message groups 
    for ( IntMsgGroupMap::iterator it=theMsgGroupIndex().begin() ; it != theMsgGroupIndex().end() ; ++it )
        (*it).second->enableMsgGroup( enable ) ; 
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */

void
CMessageGroup::disableAllMsgGroups( bool disable )
{
    // disable all message groups 
    enableAllMsgGroups( !disable ) ; 
}

/* -------------------------------------------------------------------- */

/** @param msgGroupNames Names of the message groups to enable/disable. 
 *      This list can be separated by spaces, commas or semi-colons. 
 *      An asterisk can also be used to mean all groups.
 *  @param enable Flags if the specified message groups should be enabled 
 *      or disabled.
 */

void
CMessageGroup::enableMsgGroups( const string& msgGroupNames , bool enable )
{
    // enable the specified message groups 
    const char* p = msgGroupNames.c_str() ;
    for ( ; ; )
    {
        // find the start of the next group name 
        while( isspace(*p) || *p == ',' || *p ==';' )
            p ++ ;
        if ( *p == '\0' )
            break ;

        // figure out how long the group name is 
	size_t i;
        for ( i=0 ; p[i] != '\0' ; i++ )
        {
            if ( isspace(p[i]) || p[i] == ',' || p[i] == ';' )
                break ;
        }

        // special check: if all message groups are being enabled
        if ( p[0] == '*' && p[1] == '\0' )
        {
            enableAllMsgGroups( enable ) ;
            return ;
        }

        // search for the message group name 
        CMessageGroup* pMsgGroup = findMsgGroup( string(p,i) ) ; 
        if ( pMsgGroup != NULL )
            pMsgGroup->enableMsgGroup( enable ) ; 
        // NOTE: if we didn't find the group name, just ignore it

        // move on to the next group name
        p += i ;
    }
}

/* -------------------------------------------------------------------- */

/** Dumps a table of all message groups, complete with names and descriptions
 *  where provided.
 *
 *  @param os Output stream to dump to.
 *  @param prefix Optional string to prefix each line with.
 */

void
CMessageGroup::dumpMsgGroups( ostream& os , const string& prefix )
{
    // figure out how much space to reserve for the group name 
    size_t maxLen = 0 ; 
    MsgGroupList::const_iterator it ;
    for ( it=theMsgGroupList().begin() ; it != theMsgGroupList().end() ; ++it )
    {
        size_t len = (*it)->msgGroupName().length() ;
        if ( len > maxLen )
            maxLen = len ;
    }

    // output each entry in the message group map 
    for ( it=theMsgGroupList().begin() ; it != theMsgGroupList().end() ; ++it )
    {
        // output the next message group
        const CMessageGroup* pMsgGroup = (*it) ; 
        os << prefix;
	string tmp_str( pMsgGroup->msgGroupName() );
	os << tmp_str;
	string str0(": ");
	os <<  str0; 

	string str1(" ");
        for ( size_t i=pMsgGroup->msgGroupName().length() ; i < maxLen ; i++ )
            os << str1;

        if ( ! pMsgGroup->msgGroupDescription().empty() ) 
            os << pMsgGroup->msgGroupDescription() ;
        os << endl ; 
    }
}

/* -------------------------------------------------------------------- */

CMessageGroup*
CMessageGroup::findMsgGroup( int msgGroupId )
{
    // search for the specified message group ID
    IntMsgGroupMap::iterator it = theMsgGroupIndex().find( msgGroupId ) ; 
    if ( it == theMsgGroupIndex().end() ) 
        return NULL ;
    return (*it).second ; 
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */

CMessageGroup*
CMessageGroup::findMsgGroup( const string& msgGroupName )
{
    // search for the specified message group name
    CiStringMsgGroupMap::iterator it = theMsgGroupIndexByName().find( msgGroupName ) ; 
    if ( it == theMsgGroupIndexByName().end() ) 
        return NULL ;
    return (*it).second ; 
}

/* -------------------------------------------------------------------- */

int             CMessageGroup::msgGroupId() const { return mMsgGroupId ; }
const string&   CMessageGroup::msgGroupName() const { return mMsgGroupName ; }
const string&   CMessageGroup::msgGroupDescription() const { return mMsgGroupDescription ; }
bool            CMessageGroup::isMsgGroupEnabled() const { return mIsEnabled ; }

void            CMessageGroup::enableMsgGroup( bool b ) { mIsEnabled=b ; }

/* --- INSERTERS ------------------------------------------------------ */

/// @inserters
//@{
/** @relates CMessageGroup */
 
ostream&
operator<<( ostream& os , const CMessageGroup& msgGroup )
{
    // insert the CMessageGroup
    os << "[" << msgGroup.msgGroupId() ;
    if ( ! msgGroup.msgGroupName().empty() )
        os << ":" << msgGroup.msgGroupName() ;
    os << (msgGroup.isMsgGroupEnabled() ? "on":"off") << "]" ;

    return os ; 
}

//@}
} // End namespace log
} // End namespace debug
