#ifndef __COLORRAMP_H__
#define __COLORRAMP_H__

// SYSTEM INCLUDES
//
#include <vector>

// PROJECT INCLUDES
//
#include <xmdv/RGB.h>
#include <xmdv/Visitor.h>

///
/// This class provides the interface for ColorRamp. This is used by the
/// colormap class to define the ramp to be used for colors in the map.
/// 
/// USAGE
///		cmap = new colormap(new ColorRamp(255), 255, RGB);
///		OR
///		cmap = new colormap(ColorRamp::GetPredefined1(255));
/// 
/// \date 2/4/2004
/// \author Anilkumar Patro
/// \see colormap
///
class ColorRamp
{
public:		// types

	enum RampFunction {
		CONSTANT = 0,
		LINEAR = 1,
		POLYNOMIAL = 2,
	};

	struct RampPoint
	{
		struct Point {
			float x, y;
		}			point;
		RampFunction mode;
		struct Tangent {
			float x0, y0, x1, y1;
		}			tangent;
		float		polynom[4];
	};

	typedef std::vector<RampPoint*>	PointList;
	typedef PointList::iterator		PointListIter;

public:		// methods

// LIFECYCLE
//
    ///
    /// Constructor
    /// 
    /// \param maplen size of the colormap
    ///
	ColorRamp(unsigned int maplen);
    ///
    /// Constructor - initialize with a predefined set of ramp points
    /// 
    /// \param maplen size of the colormap
    /// \param N number of ramp points
    /// \param list array of ramp points
    ///
	ColorRamp(unsigned int maplen, unsigned int N, float list[][5]);
    ///
    /// Copy Constructor
    /// 
    /// \param ColorRamp& copy this
    ///
	ColorRamp(const ColorRamp&);
    ///
    /// Destructor
    /// 
    /// \param void 
    ///
	~ColorRamp(void);

	static ColorRamp *GetPredefined1(int maplen);
	static ColorRamp *GetPredefined2(int maplen);
	static ColorRamp *GetPredefinedGreyScale(int maplen);

// OPERATORS
//
    ///
    /// Assignment operator
    /// 
    /// \param ColorRamp& 
    /// \return void operator  
    /// \pre 
    /// \post 
    ///
	void operator = (const ColorRamp&);

// OPERATIONS
//
	void Accept(Visitor<ColorRamp>& visitor);

    ///
    /// Initialize the ColorRamp using a list of points defining the ramp
    /// 
    /// \param N number of points
    /// \param list array of points
    ///
	void Initialize(unsigned int N, float list[][5]);

    ///
    /// Evaluate the value of the ramp at a particular position and for 
	/// the specified component of the color
    /// 
    /// \param index position of point
    /// \param cmpnt component of the color
    /// \return float value at the position on the ramp
    ///
	float Evaluate(int index, int cmpnt);

	PointListIter Begin(int cmpnt);
	PointListIter End(int cmpnt);

    ///
    /// Insert a point on the ramp
    /// 
    /// \param x X position
    /// \param y Y position
    /// \param ndx Index value
    /// \param cmpnt Component of the color
    /// \param point A pointer to the new ramp point
    /// \return int 
    ///
	int InsertPoint(float x, float y, 
					int ndx, int cmpnt, PointListIter &point);

    ///
    /// Remove a point from the ramp
    /// 
    /// \param iter the pointer to the point to be removed
    /// \param cmpnt from this component of the color
    ///
	void RemovePoint(PointListIter iter, int cmpnt);

	///
	/// Move a point on the ramp to any position
	/// 
	/// \param curPoint
	/// \param curCmpnt
	/// \param x
	/// \param y
	/// \return
	/// 
	int MovePoint(PointListIter curPoint, int curCmpnt,	
				  float x, float y);

    ///
    /// Move the tangent of a particular point on the ramp
    /// 
    /// \param curPoint 
    /// \param curCmpnt 
    /// \param sens 
    /// \param x 
    /// \param y 
    /// \return int  
    ///
	int MoveTangent(PointListIter curPoint, int curCmpnt, int sens, 
					float x, float y);

	///
	/// Update the lookup table
	/// 
	/// \param cmpnt Component of color
	/// \param curr Point to be considered
	///
	void Update(int cmpnt, PointListIter curr);

	void Invert();
	void Reverse();

private:	// members

	int		m_dimlut;		//< size of the lookup table
	int		m_maxndx;		//< max addressable index values in the table
	float	m_lutratio;		//< maxndx / 255
	float	m_tanlen;		//< length of the tangent points from the point
	float	*m_tabcmpnt;	//< the lookup table

	PointList m_userPoints[4]; //< the four component valued points

private:	// methods

	///
	/// Retrieve the coefficients to make polynomial function
	/// 
	/// \param mode function to be used
	/// \param a0,b0,t1 
	/// \param a1,b1,t1
	/// \param coeff
	/// 
	void GetPolynomial(RampFunction mode,
					   float a0, float b0, float t0,
					   float a1, float b1, float t1,
					   float coeff[4]);

    ///
    /// Solve a polynomial function given by its coefficient values by 
	/// substituting the given x value
    /// 
    /// \param x value
    /// \param *coeff coefficients in the polynomial
    /// \return float evaluated value
    ///
	float SolvePolynomial(float x, float *coeff);

};

//-------------------------------------------------------------------------
inline void ColorRamp::Accept(Visitor<ColorRamp>& visitor)
{
	visitor.Visit(*this);
}
//-------------------------------------------------------------------------

#endif //__COLORRAMP_H__
