#include <testcache/Printer.h>
#include <fstream>
#include <cache/ConfigurationStore.h>

#pragma warning(disable: 4786 4250 4503 4244)
#include <DOM/SAX2DOM/SAX2DOM.h>
#include <SAX/helpers/CatchErrorHandler.h>
#include <DOM/Utils/Stream.h>

PrintFileParser::PrintFileParser(std::string fileName,
								 cache::CacheConfigStore* configStore,
								 cache::DBConfigStore* dbConfigStore)
{
	//Checking to see if the file exists.
	std::ifstream inputStream(fileName.c_str(),std::ios::in);
	if(!inputStream.good())
	{
		std::cout<<"Print File Not found "<<fileName;		
	}
	inputStream.close();
	this->cacheConfigStore = configStore;
	this->dbConfigStore = dbConfigStore;
	////////////////////////////////////////////

	SAX2DOM::Parser<std::string> domParser;
	SAX::CatchErrorHandler<std::string> eh;
	domParser.setErrorHandler(eh);	
	SAX::InputSource is;
	is.setSystemId(fileName);		
	domParser.parse(is);	
	DOM::Document<std::string> doc = domParser.getDocument();
	printerMap = new std::map<std::string,Printer*>;
	makePrinters(doc);
}


void PrintFileParser::makePrinters(DOM::Document<std::string>& doc)
{
	DOM::NodeList<std::string> nodeList =  doc.getElementsByTagName("CacheSize");

	Printer *p = new CacheSizePrinter(this->dbConfigStore,
					this->cacheConfigStore,
					nodeList);

	(*printerMap)[p->getID()] = p;

	nodeList = doc.getElementsByTagName("InputUserTrace");
	p = new InputUserTracePrinter(this->dbConfigStore,
					this->cacheConfigStore,
					nodeList);
	(*printerMap)[p->getID()] = p;

	
	nodeList = doc.getElementsByTagName("ReplacementPolicy");
	p = new ReplacementPolicyPrinter(this->dbConfigStore,
									this->cacheConfigStore,
									nodeList);
	(*printerMap)[p->getID()] = p;


	nodeList = doc.getElementsByTagName("SecondaryIndex");
	p = new SecondaryIndexPrinter(this->dbConfigStore,
								  this->cacheConfigStore,
								  nodeList);
	(*printerMap)[p->getID()]= p;

	nodeList = doc.getElementsByTagName("MainMemoryIndex");
	p = new MainMemoryIndexPrinter(this->dbConfigStore,
						this->cacheConfigStore,
						nodeList);
	(*printerMap)[p->getID()]= p;


	nodeList = doc.getElementsByTagName("Prefetcher");
	p = new PrefetchPrinter(this->dbConfigStore,
						this->cacheConfigStore,
						nodeList);
	(*printerMap)[p->getID()]= p;
	

	//outputstatistics.
	nodeList = doc.getElementsByTagName("TotalTuplesFetched");
	p = new TotalTuplesPrinter(this->dbConfigStore,
							this->cacheConfigStore,
							nodeList);
	(*printerMap)[p->getID()]= p;


	nodeList = doc.getElementsByTagName("TotalLatency");
	p = new TotalLatencyPrinter(this->dbConfigStore,
						this->cacheConfigStore,
						nodeList);
	(*printerMap)[p->getID()]= p;


	nodeList = doc.getElementsByTagName("TotalSearchLatency");
	p = new TotalSearchLatencyPrinter(this->dbConfigStore,
								this->cacheConfigStore,
								nodeList);	
	(*printerMap)[p->getID()]= p;
	

	nodeList = doc.getElementsByTagName("Latency");
	p = new AvgLatencyPrinter(this->dbConfigStore,
						this->cacheConfigStore,
						nodeList);
	(*printerMap)[p->getID()]= p;

	nodeList = doc.getElementsByTagName("MedianLatency");
	p = new MedianLatencyPrinter(this->dbConfigStore,
					this->cacheConfigStore,
					nodeList);
	(*printerMap)[p->getID()]= p;


	nodeList = doc.getElementsByTagName("FragmentationThreshold");
		p = new FragThresholdPrinter(this->dbConfigStore,
									this->cacheConfigStore,
									nodeList);
	(*printerMap)[p->getID()]= p;

	nodeList = doc.getElementsByTagName("HitRatio");
		p = new HitRatioPrinter(this->dbConfigStore,
				this->cacheConfigStore,
				nodeList);
	(*printerMap)[p->getID()]= p;
}
/**
 * Retunrs the map that contains id and the name
 * of all the Printers
 * \return 
 */
std::map<std::string,Printer*>* PrintFileParser:: getMap()
{
	return(this->printerMap);
}

Printer::Printer(cache::DBConfigStore* dbConfigStore,
				 cache::CacheConfigStore *cacheConfigStore,
				 DOM::NodeList<std::string>& nodeList)
{
	this->isEnabled = false;
	DOM::Node<std::string> node = nodeList.item(0);
	DOM::NamedNodeMap<std::string> nameMap =  node.getAttributes();
	std::string en = nameMap.getNamedItem("print").getNodeValue();
	if(en == "on")
	{
		this->isEnabled = true;
	}
}

Printer::Printer(cache::DBConfigStore* dbConfigStore,
				 cache::CacheConfigStore *cacheConfigStore,
				 DOM::Node<std::string>& node)
{	
	this->isEnabled = false;
	DOM::NamedNodeMap<std::string> nameMap =  node.getAttributes();
	std::string nodeName = node.getNodeName();
	std::string en = nameMap.getNamedItem("print").getNodeValue();
	if(en == "on")
	{
		this->isEnabled = true;
	}
}

void Printer::printName(std::ostream &outputStream)
{
	if(this->isEnabled)
		outputStream<<this->id<<"\t";
}

std::string Printer::getID()
{
	return(this->id);
}

void Printer::setValue(std::string value){}

void Printer::setValue(double value){}

void Printer::setValue(unsigned long value){}

void Printer::setValue(unsigned int value){}

void Printer::setValue(int value){}

std::ostream& operator <<(std::ostream& outputStream,Printer& printer)
{
	if(printer.isEnabled)
		printer.printValue(outputStream);

	return(outputStream);
}


CacheSizePrinter::CacheSizePrinter(cache::DBConfigStore* dbConfigStore,
								   cache::CacheConfigStore *cacheConfigStore,
								   DOM::NodeList<std::string>& nodeList):
								   Printer(dbConfigStore,cacheConfigStore,nodeList)
{	
	this->value = cacheConfigStore->getCacheSize();
	this->id = "CacheSize";
}

void CacheSizePrinter::printValue(std::ostream &outputStream)
{
	if(this->isEnabled)
	outputStream<<value<<"\t";	
}

InputUserTracePrinter  ::InputUserTracePrinter (cache::DBConfigStore* dbConfigStore,
												cache::CacheConfigStore *cacheConfigStore,
												DOM::NodeList<std::string>& nodeList):
												Printer(dbConfigStore,cacheConfigStore,nodeList)
{
	this->id = "InputUserTrace";
}

void InputUserTracePrinter  ::printValue(std::ostream &outputStream)
{
	if(this->isEnabled)
		outputStream<<this->value<<"\t";
}

void InputUserTracePrinter ::setValue(std::string inputUserTrace)
{
	this->value =inputUserTrace;	
}


ReplacementPolicyPrinter ::ReplacementPolicyPrinter (cache::DBConfigStore* dbConfigStore,
												cache::CacheConfigStore *cacheConfigStore,
												DOM::NodeList<std::string>& nodeList):
			Printer(dbConfigStore,cacheConfigStore,nodeList)
{
	
	this->value = cacheConfigStore->getReplacementPolicy();
	this->id = "ReplacementPolicy";
}

void ReplacementPolicyPrinter ::printValue(std::ostream &outputStream)
{
	if(this->isEnabled)
		outputStream<<this->value<<"\t";
}



SecondaryIndexPrinter::SecondaryIndexPrinter(cache::DBConfigStore* dbConfigStore,
												cache::CacheConfigStore *cacheConfigStore,
												DOM::NodeList<std::string>& nodeList):
												Printer(dbConfigStore,cacheConfigStore,nodeList)
{
	
	this->value = cacheConfigStore->queryType;
	this->id = "SecondaryIndex";
}

void SecondaryIndexPrinter::printValue(std::ostream &outputStream)
{
	if(this->isEnabled)
		outputStream<<this->value<<"\t";
}

MainMemoryIndexPrinter::MainMemoryIndexPrinter(cache::DBConfigStore* dbConfigStore,
											 cache::CacheConfigStore *cacheConfigStore,
											 DOM::NodeList<std::string>& nodeList):
Printer(dbConfigStore,cacheConfigStore,nodeList)
{	
	this->children = NULL;
	this->id = "MainMemoryIndex";
	this->value = cacheConfigStore->memoryIndex;	
	std::string value = nodeList.item(0).getNodeName();
	DOM::Node<std::string> node = nodeList.item(0);	
	DOM::NodeList<std::string> childNodes = node.getChildNodes();
	
	if(childNodes != NULL)
	{		
		this->children = new std::vector<Printer*>;
		
		DOM::Node<std::string> node = childNodes.item(1);
		RTreeNodeSizePrinter* rtnszp = new RTreeNodeSizePrinter(dbConfigStore,
											cacheConfigStore,
											node);
		this->children->push_back(rtnszp);


		DOM::Node<std::string> node3 = childNodes.item(3);
		RTreeIndexPrinter* rtnixp = new RTreeIndexPrinter(dbConfigStore,
										cacheConfigStore,
										node3);
		this->children->push_back(rtnixp);

		DOM::Node<std::string> node5 = childNodes.item(5);
		RTreeFillFactorPrinter* rtffp = new RTreeFillFactorPrinter(dbConfigStore,
											cacheConfigStore,
											node5);
		this->children->push_back(rtffp);
	} // if(children != NULL)	
}

void MainMemoryIndexPrinter::printValue(std::ostream &outputStream)
{
	if(this->isEnabled)
		outputStream<<this->value<<"\t";
	if(this->children != NULL)
	{
		for(unsigned int i = 0;i<children->size();i++)
		{
			(*children)[i]->printValue(outputStream);
		}
	}
}

MainMemoryIndexPrinter::~MainMemoryIndexPrinter()
{
	for(unsigned int i = 0;i<children->size();i++)
	{
		Printer* printer = (*children)[i];		
		delete(printer);
	} // while(children->size() != 0)
	children->clear();
	delete(children);
}


PrefetchPrinter::PrefetchPrinter(cache::DBConfigStore* dbConfigStore,
											 cache::CacheConfigStore *cacheConfigStore,
											 DOM::NodeList<std::string>& nodeList):
Printer(dbConfigStore,cacheConfigStore,nodeList)
{
	this->id = "Prefetcher";
	this->value = cacheConfigStore->getPrefetcherType();
	if(cache::CacheConfigStore::Noprefetch == cacheConfigStore->getPrefetchStrategy())
		this->value = "Disabled";
	else
	if(cache::CacheConfigStore::Adaptive== cacheConfigStore->getPrefetcherType())
		this->value = "Adaptive";
	else
	if(cache::CacheConfigStore::Fixed== cacheConfigStore->getPrefetcherType())
		this->value = "Fixed";
}


void PrefetchPrinter::printValue(std::ostream &outputStream)
{
	if(this->isEnabled)
		outputStream<<this->value<<"\t";
}


TotalTuplesPrinter::TotalTuplesPrinter(cache::DBConfigStore* dbConfigStore,
								 cache::CacheConfigStore *cacheConfigStore,
								 DOM::NodeList<std::string>& nodeList):
Printer(dbConfigStore,cacheConfigStore,nodeList)
{
	
	this->id = "TotalTuplesFetched";
}

void TotalTuplesPrinter ::setValue(unsigned long totalTuples)
{
	this->value = totalTuples;
}

void TotalTuplesPrinter::printValue(std::ostream &outputStream)
{
	if(this->isEnabled)
		outputStream<<this->value<<"\t";
}

TotalLatencyPrinter::TotalLatencyPrinter(cache::DBConfigStore* dbConfigStore,
								 cache::CacheConfigStore *cacheConfigStore,
								 DOM::NodeList<std::string>& nodeList):
								 Printer(dbConfigStore,cacheConfigStore,nodeList)
{
	
	this->id = "TotalLatency";
}

void TotalLatencyPrinter::setValue(unsigned long latency)
{
	this->value =latency;	
}

void TotalLatencyPrinter::printValue(std::ostream &outputStream)
{
	if(this->isEnabled)
		outputStream<<this->value<<"\t";
}

TotalSearchLatencyPrinter::TotalSearchLatencyPrinter(cache::DBConfigStore* dbConfigStore,
						  cache::CacheConfigStore *cacheConfigStore,
						  DOM::NodeList<std::string>& nodeList):
Printer(dbConfigStore,cacheConfigStore,nodeList)
{
	this->id = "TotalSearchLatency";
}

void TotalSearchLatencyPrinter ::printValue(std::ostream &outputStream)
{
	if(this->isEnabled)
		outputStream<<this->value<<"\t";
}

void TotalSearchLatencyPrinter ::setValue(unsigned long totalSearchLatency)
{
	this->value = totalSearchLatency;
}


AvgLatencyPrinter::AvgLatencyPrinter(cache::DBConfigStore* dbConfigStore,
						  cache::CacheConfigStore *cacheConfigStore,
						  DOM::NodeList<std::string>& nodeList):
Printer(dbConfigStore,cacheConfigStore,nodeList)
{
	this->id = "AvgLatency";
}

void AvgLatencyPrinter::printValue(std::ostream &outputStream)
{
	if(this->isEnabled)
		outputStream<<this->value<<"\t";
}

void AvgLatencyPrinter::setValue(double avgLatency)
{
	this->value = avgLatency;
}

MedianLatencyPrinter::MedianLatencyPrinter(cache::DBConfigStore* dbConfigStore,
										cache::CacheConfigStore *cacheConfigStore,
										DOM::NodeList<std::string>& nodeList):
Printer(dbConfigStore,cacheConfigStore,nodeList)
{
	this->id = "MedianLatency";
}


void MedianLatencyPrinter::printValue(std::ostream &outputStream)
{
	if(this->isEnabled)
		outputStream<<this->value<<"\t";
}

void MedianLatencyPrinter ::setValue(double medianLatency)
{
	this->value = medianLatency;
}

void setValue(double medianLatency);	

HitRatioPrinter::HitRatioPrinter(cache::DBConfigStore* dbConfigStore,
									 cache::CacheConfigStore *cacheConfigStore,
									 DOM::NodeList<std::string>& nodeList):
Printer(dbConfigStore,cacheConfigStore,nodeList)
{
	this->id = "HitRatio";
}

void HitRatioPrinter::printValue(std::ostream &outputStream)
{
	if(this->isEnabled)
		outputStream<<this->value<<"\t";
}

void HitRatioPrinter::setValue(double hitRatio)
{
	this->value = hitRatio;
}

FragThresholdPrinter::FragThresholdPrinter(cache::DBConfigStore* dbConfigStore,
								 cache::CacheConfigStore *cacheConfigStore,
								 DOM::NodeList<std::string>& nodeList):
Printer(dbConfigStore,cacheConfigStore,nodeList)
{
	this->id = "FragThreshold";
}

void FragThresholdPrinter::printValue(std::ostream &outputStream)
{
	if(this->isEnabled)
		outputStream<<this->value<<"\t";
}

void FragThresholdPrinter::setValue(unsigned int fragThreshold)
{
	this->value = fragThreshold;
}


RTreeNodeSizePrinter::RTreeNodeSizePrinter(cache::DBConfigStore* dbConfigStore,
											cache::CacheConfigStore *cacheConfigStore,
											DOM::Node<std::string>& node):
Printer(dbConfigStore,cacheConfigStore,node)
{
	this->id = "RTreeNodeSize";
	this->value = cacheConfigStore->getRTreeLeafCapacity();
}

void RTreeNodeSizePrinter ::printValue(std::ostream &outputStream)
{
	if(this->isEnabled)
		outputStream<<value<<"\t";
}

RTreeIndexPrinter::RTreeIndexPrinter(cache::DBConfigStore* dbConfigStore,
										   cache::CacheConfigStore *cacheConfigStore,
										   DOM::Node<std::string>& node):
Printer(dbConfigStore,cacheConfigStore,node)
{
	this->id = "RTreeIndexPrinter";
	this->value = cacheConfigStore->getRTreeIndexCapacity();
}

void RTreeIndexPrinter ::printValue(std::ostream &outputStream)
{
	if(this->isEnabled)
		outputStream<<value<<"\t";
}

RTreeFillFactorPrinter::RTreeFillFactorPrinter(cache::DBConfigStore* dbConfigStore,
										   cache::CacheConfigStore *cacheConfigStore,
										   DOM::Node<std::string>& node):
Printer(dbConfigStore,cacheConfigStore,node)
{
	this->id = "RTreeFillFactorPrinter";
	this->value = cacheConfigStore->getRTreeFillFactor();
}

void RTreeFillFactorPrinter ::printValue(std::ostream &outputStream)
{	
	if(this->isEnabled)
		outputStream<<value<<"\t";
}


