#include <algorithm>
#include <xmdv/BrushMgr.h>
#include <xmdv/brush_oper.h>
#include <xmdv/okcdata.h>
#include <xmdv/XmdvTool.h>

#include <iostream>
#include <float.h>

// Used for sorting
// Anonymous namespaces don't clutter up the global namespace
namespace 
{
	// This temporary buffer of data is a copy of the data of all
	//  lines at a particular dimension d. This is temporarily used
	//  to sort Index[d] because it would be inefficient to randomly
	//  access double over a spread of (in the case of large files)
	//  hundreds of megabytes.
	double *	__tmp_points;

	// This function is used for std::srt as the predicate because
	// an inline function is much faster than a stl-style
	// predicate class's operator()(int, int).
	inline	int __SortPredicate(int lhs, int rhs)
	{	return __tmp_points[lhs] < __tmp_points[rhs];	}
}

BrushMgr::BrushMgr(okcdata & owner)
: Dataset(owner)
{
	/********************** Used for sorting **********************/
	// This assumes that okcdata has loaded all of its data by now.
	Index.SetLen(Dataset.all_dims);

	::__tmp_points = new double[Dataset.data_size];

	// First, find out which lines are out of bounds
	// It is unecessary to store other information about those lines.
  	OutOfBounds.SetLen(Dataset.data_size);

	// Set this to data_size, and decrement for every line that ends up
	// being out of bounds in the for() loop.
	ValidLineCount = Dataset.data_size;

	for (int l = 0; l < Dataset.data_size; ++l)
	{
        // 'data' is a direct pointer into the data buffer of the okcdata
		// so that data[0..all_dims-1] contains the data of a line
		double * data = Dataset.data_buf.addr() + Dataset.all_dims * l;
		bool & thisMark = OutOfBounds[l];

		for (int d = 0; d < Dataset.all_dims; ++d)
		{
			thisMark = (data[d] < Dataset.cg_dim_min[d]) || (data[d] > Dataset.cg_dim_max[d]);
			if (thisMark)
			{
				ValidLineCount--;
				break;	// Leave early
			}
		}
	}

	ARR<int> sorted(Dataset.data_size);

	for (int d = 0; d < Dataset.all_dims; ++d)
	{
		if (Dataset.cg_dim_min[d] == Dataset.cg_dim_max[d])
		{
			// This dimension does not need an index because all of
			// the data on this dimension is the same (assuming that
			// the file's min and max values are accurate).

			// If an index is not present for a dimension d, this is
			// made clear by the fact that Index[d] has a length of 0.

            // Index[d] is left alone, and its implicit arg-less ctor
			// leaves it with a length of 0.
			continue;
		}

		// Load line data at a dimension d
		ARR<int> & thisIndex = Index[d];
		thisIndex.SetLen(ValidLineCount);
		
		// Index into Dataset.data_buf to select a line's value at
		//  dimension d
		int buf_index = d;
		int growby = Dataset.all_dims;

		// Create the unsorted list of indices and copy the data
		for (int l = 0; l < Dataset.data_size; ++l, buf_index += growby)
		{
			// Set every line's index to its position in the file
			sorted[l] = l;
			// Keep out of bounds lines out of this
			if (OutOfBounds[l])
			{
				// Set this line's number to something unreasonably high, so that
				// lines that are out of bounds will sink to the farthest end of
				// the array, so that they can be ignored.
				::__tmp_points[l] = DBL_MAX;
			}
			else
			{
				::__tmp_points[l] = Dataset.data_buf[buf_index];
			}
		}

		// Sort the indices for this dimension, given *this as the
		// predicate for the STL quicksort algorithm. Pointers
		// work like STL iterators, so this code is correct, even
		// though the second pointer is outside of the valid range
		// of pointers to thisIndex
		std::sort(	sorted.addr(0),
					sorted.addr(sorted.Len()),
					::__SortPredicate);

		// Does the min equal the max?
		if (__tmp_points[sorted[0]] == __tmp_points[sorted[ValidLineCount-1]])
		{
			// It turns out this dimension has no variance (constant),
			// so free the index data for this dimension and ignore.
            thisIndex.Clear();
			continue;
		}

		for (int i = 0; i < ValidLineCount; ++i)
		{
			// Indices of invalid lines are at the far right of the array
			thisIndex[i] = sorted[i];
		}

		// thisIndex is now an array of integers that correspond to
		// the indices of data lines sorted for dimension d
	}

	// Now that the indices have been created, we can now efficiently
	// calculate the contents of each dimension. This is continued in
	// BrushMgr::ResetBrushCounter(), but the arrays can still be sized.
	for (int b = 0; b < MAXBRUSH; ++b)
		BrushCounters[b].Set(Dataset.data_size, Dataset.dims);

	// Free temp data
	delete [] ::__tmp_points;
	__tmp_points = 0;
}

void BrushMgr::FormatDataValuesLine(int line, char * str)
{
	char buf[100];

	int oplines = 0;
	for (int i = 0; i < num_oper; ++i)
	{
		if (brush_ops[i].oper & VALUES)
			oplines += BrushOpValues[i].Len();

		if ((brush_ops[i].oper & AVERAGE) && BrushOpValues[i].Len())
			oplines++;
	}

	// Either the header line or one of the separators?
	if (line == 0 || line == 1 || line == oplines + 2)
	{
        int i;
		strcpy(str, "Oper: ");
		for (i = 0; i < Dataset.dims; i++)
		{
			sprintf(buf, " %12s", Dataset.names[i]);
			strcat(str, buf);
		}

		// A seperator
		if (line != 0)
		{
			// Reformat with '-'
			i = 0;
			while (str[i] != 0)
				str[i++] = '-';
		}
	}
	// Summary line?
	else if (line == oplines + 3)
	{
		int count = 0;
		for (int op = 0; op < num_oper; ++op)
			if (brush_ops[op].oper & VALUES)
				count += BrushOpValues[op].Len();
		sprintf(str, "Total number of brushed data points = %d", count);
	}
	else
	{
		ARR< double> data(Dataset.dims);
		int counter = line - 2;	// The first 2 lines are the labels and the seperator
		// Find which op number this line includes
		int op_number;
		for (op_number = 0; op_number < num_oper; ++op_number)
		{
			// Consider a line value?
			if (brush_ops[op_number].oper & VALUES)
			{
				// Is the line within the range of lines inside the brush?
				if (counter < BrushOpValues[op_number].Len())
				{
					sprintf(str, "%4d: ", op_number + 1);
					// GetData after you know you'll need it. After.
					Dataset.GetData(data, BrushOpValues[op_number][counter]);
					for (int j = 0; j < Dataset.dims; j++)
					{
						sprintf(buf, " %12.3f", data[j]);
						strcat(str, buf);
					}
					return;
				}
				else
					// Skip the values in this brush
					counter -= BrushOpValues[op_number].Len();
			}

			// Consider an average?
			if (brush_ops[op_number].oper & AVERAGE)
			{
				// Only print information if 
				if (counter == 0 && BrushOpValues[op_number].Len() != 0)
				{
					// This line should have average information due to
					// the fact that we checked the contents of the operation
					//BrushOperCompAverage(op_number, data);
					ARR<double> & avg = BrushOpAverages[op_number];
					sprintf(str, "Avg%d: ", op_number + 1);
					for (int j = 0; j < g_globals.dataset->dims; j++)
					{
						//sprintf(buf, " %12.3f", data[j]);
						sprintf(buf, " %12.3f", avg[j]);
						strcat(str, buf);
					}
					// Done
					return;
				}
				else
					counter--; // Move ahead 1 line
			}
		}
	}
}

// PAUL: tk_opt
// This function takes the fraction of lines above the top of the visible text,
// the fraction of the text above the bottom of the visible text, and the number of
// lines in the text widget in order to compute which lines of the DataValuesTest
// widget need to be formatted.
int BrushMgr::BrushedDataValuesUpdate(ClientData clientData, Tcl_Interp* interp,
										int argc, char* argv[])
{
	if (!g_globals.dataset)
		return TCL_OK;

	if (argc != 4)
	{
		return TCL_ERROR;
	}

	using namespace std;
	
	// Get the range of lines that need to be updated.
	int linecount = atoi(argv[3]);
	int lower = (int)floor(atof(argv[1]) * linecount);
	int upper = (int)ceil(atof(argv[2]) * linecount);
	if (upper == linecount - 1)
		upper--;
	else if (upper == linecount)
		upper -= 2;

	//cout << "[] = " << lower << ", " << upper << endl;

	char str[50000];
	char tcl_command[50000];

	//Tcl_VarEval(interp, "[W DataValuesText] conf -state normal", 0);
	for (int line = lower; line <= upper; ++line)
	{
		// Print the contents of individual lines
		sprintf(tcl_command, "[W DataValuesText] delete %d.0 %d.end", line+1, line+1);
		Tcl_Eval(interp, tcl_command);

		str[0] = 0;
		g_globals.dataset->brushMgr->FormatDataValuesLine(line, str);

		sprintf(tcl_command, "[W DataValuesText] insert %d.0 \"%s\"", line+1, str);
		Tcl_Eval(interp, tcl_command);
	}
	//Tcl_VarEval(interp, "[W DataValuesText] conf -state disabled", 0);
	return TCL_OK;
}

void BrushMgr::ResetBrushCounters(int b)
{
	// By default, a line is outside of a brush
	// as many times as there dimensional ranges
	// to be outside of. Later, the counter for
	// each line will be decreased for each range
	// each line is within.
	const int default_count = Dataset.dims;

	// One counter per line (per brush)
	for (int l = 0; l < Dataset.data_size; ++l)
		// Note that OutOfBounds lines will not change
		BrushCounters[b][l] = default_count;

	Brush & thisBrush = g_globals.brushes[b];//*/all_brushes[b];

	// Bounds of the brush at a dimension d
	int First, Last = 0;
	for (int d = 0, count = 0; d < Dataset.all_dims; ++d)
	{
		/*
			The enumeration of dimensions is tricky, due to the
			fact that the brush, the on[] array, and the file
			all have different methods of locating information
			by dimension index.

			- d enumerates dimensions as they are shown in the dialog
			called "Dimension On/Off/Reorder." This means that it
			includes hidden dimensions. okcdata::on is ordered in
			this fashion.

			- brush_d enumerates dimensions in view-space (the order that
			dimensions are shown to the screen). It is similar to 'd'
			except that hidden dimensions are skipped.

			- count is used to count the number of unhidden dimensions
			we have run enumerated.

			- file_d is the index of information by its location in
			the file. It does not change as dimensions are reorderd
			or hidden from view.
		*/
		if (!Dataset.on[d])
			continue;

		// Get the absolute dim index
		int file_d = Dataset.map_buf_to_file[d];
		int brush_d = count;
		count++;

		// Constant dimension (all or nothing)
		if (Index[file_d].Len() == 0)
		{
			// Are ALL lines within the brush range?
			if (Dataset.cg_dim_min[file_d] >= thisBrush.Min(brush_d) &&
				Dataset.cg_dim_min[file_d] <= thisBrush.Max(brush_d))
			{
				// The brush contains all lines at this dimension
				// Just set all valid lines a bit lower
				for (int i = 0; i < Dataset.data_size; ++i)
					if (!IsLineOutOfBounds(i))
						BrushCounters[b][i]--;
			}
		}
		else if ((First = SearchLowerBound(file_d, thisBrush.Min(brush_d))) != -1 && 
				 (Last  = SearchUpperBound(file_d, thisBrush.Max(brush_d))) != -1)
		{
			// Use a direct pointer to the array
			const int * thisIndex = Index[file_d].addr(First);

            // Dont enter if the range [First, Last] is not valid
			for (int i = First; i <= Last; ++i)
			{
				// Lines within the interval of the index are
				// within the brush, so we better mark them.
				BrushCounters[b][*(thisIndex++)]--;
			}
		}
	}

	// Now, each line l is inside of brush b 
	//	iff BrushCounter[b][l] == 0

	BrushOpValues.Clear();
	BrushOpValues.SetLen(num_oper);
	BrushOpAverages.Clear();
	BrushOpAverages.SetLen(num_oper);

	// The intermediate buffer of size ValidLineCount
	// ValidLineCount is big enough because no invalid
	// lines can qualify as being in a brush
	ARR<int> IndexBuffer(ValidLineCount);

	// Recompute the contents of the brush ops
	for (int op = 0; op < num_oper; op++)
	{
		// The number of lines in this operation
		int count = 0;

		for (int l = 0; l < Dataset.data_size; ++l)
		{
			if (IsLineInBrushOper(l, op))
			{
				IndexBuffer[count++] = l;
			}
		}

		// Copy the contents
		BrushOpValues[op].SetLen(count);
		for (int i = 0; i < count; ++i)
			BrushOpValues[op][i] = IndexBuffer[i];
		
		// Average?
		if ((brush_ops[op].oper & AVERAGE) && (BrushOpValues[op].Len()))
		{
			// Init the average array to the number of dimensions
			BrushOpAverages[op].Set(Dataset.dims, 0.0f);
			// Clearer access to average
			ARR<double> & thisAvg = BrushOpAverages[op];
			// Container
			ARR<double> data(Dataset.dims);
			for (int i = 0; i < BrushOpValues[op].Len(); ++i)
			{
				Dataset.GetData(data, BrushOpValues[op][i]);
				for (int d = 0; d < Dataset.dims; ++d)
					thisAvg[d] += data[d];
			}
			double inv_count = 1. / BrushOpValues[op].Len();
			for (int d = 0; d < Dataset.dims; ++d)
				thisAvg[d] *= inv_count;
		}
	}
}


bool BrushMgr::IsLineInBrushOper(int line, int br_op) const
{
	BrushOperation * op = &brush_ops[br_op];
	for (int i = 0; i < op->num_brush; ++i)
	{
		int brush_num = op->brush[i];
		if (!g_globals.brushes[brush_num].enable)
			return false;
	}

	bool result = IsLineInBrush(line, op->brush[0]);
	if (op->xnot[0])
		result = !result;

	for (int i = 1; i < op->num_brush; ++i)
	{
		bool next_br = IsLineInBrush(line, op->brush[i]);
		if (op->xnot[i])
			next_br = !next_br;

        switch(op->conj[i-1])
        {
        case AND:
            result &= next_br;
            break;
        case OR:
            result |= next_br;
            break;
        case XOR:
            result ^= next_br;
            break;
        }
	}
	return result;
}

void BrushMgr::DisableDim(int d)
{
	// Decrease the counter for all lines that are outside of each 
	// brush at dimension d, since d is to be hidden. This way,
	// the hidden dimension has no bearing on brush inclusion.
	for (int b = 0; b < MAXBRUSH; ++b)
	{
		Brush & thisBrush = g_globals.brushes[b];//all_brushes[b];

		// Constant dimension (all or nothing)
		if (Index[d].Len() == 0)
		{
			// Are NONE of the lines within the brush range?
			if (Dataset.cg_dim_min[d] < thisBrush.Min(d) ||
				Dataset.cg_dim_min[d] > thisBrush.Max(d))
			{
				// The brush contains none lines at this dimension
				// Just set all valid lines a bit lower
				for (int i = 0; i < Dataset.data_size; ++i)
					if (!IsLineOutOfBounds(i))
						BrushCounters[b][i]--;
			}
		}
		else
		{
			int First = SearchLowerBound(d, thisBrush.Min(d));
			int Last  = SearchUpperBound(d, thisBrush.Max(d));

			// Decrement lines OUTSIDE of the range of the brush
			if (First != -1) for (int i = 0; i < First; ++i)
			{
				BrushCounters[b][Index[d][i]]--;
			}
			if (Last != -1) for (int i = Last + 1; i < Index[d].Len(); ++i)
			{
				BrushCounters[b][Index[d][i]]--;
			}
		}
	}
}

void BrushMgr::EnableDim(int d)
{
//	printf("EnableDim %d\n", d);

	// Increase the counter for all lines that are outside of each 
	// brush at dimension d, since d is to be enabled
	for (int b = 0; b < MAXBRUSH; ++b)
	{
		Brush & thisBrush = g_globals.brushes[b];//all_brushes[b];

		// Constant dimension (all or nothing)
		if (Index[d].Len() == 0)
		{
			// Are NONE of the lines within the brush range?
			if (Dataset.cg_dim_min[d] < thisBrush.Min(d) ||
				Dataset.cg_dim_min[d] > thisBrush.Max(d))
			{
				// The brush contains none lines at this dimension
				// Just set all valid lines a bit higher
				for (int i = 0; i < Dataset.data_size; ++i)
					if (!IsLineOutOfBounds(i))
						BrushCounters[b][i]++;
			}
		}
		else
		{
			int First = SearchLowerBound(d, thisBrush.Min(d));
			int Last  = SearchUpperBound(d, thisBrush.Max(d));

			// Increment lines OUTSIDE of the range of the brush
			if (First != -1) for (int i = 0; i < First; ++i)
			{
				BrushCounters[b][Index[d][i]]++;
			}
			if (Last != -1) for (int i = Last + 1; i < Index[d].Len(); ++i)
			{
				BrushCounters[b][Index[d][i]]++;
			}
		}
	}
}

void BrushMgr::ResetLineColors()
{
	// For each line
	for (int l = 0; l < g_globals.dataset->data_size; l++)
	{
		okcdata::mark_type & mark = Dataset.mark[l];

		// Keep out of bounds lines out of this
		if (OutOfBounds[l])
		{	
			mark = okcdata::DISABLED;
			continue;
		}

		RGBt & color = Dataset.color[l];
		mark = okcdata::REGULAR;

		// For each operation
		for (int op = num_oper-1; op >= 0; --op)
		{
			if (brush_ops[op].num_brush == 0) continue;

			bool enclosed = IsLineInBrushOper(l, op);
			if (!enclosed) continue;

			if (brush_ops[op].oper & HIGH)
			{
				mark = okcdata::COLORED;

				// NO RAMP COLORING HERE
				// ...
				color = interface_color[HIGHLIGHT1+op];
			}
			else if (brush_ops[op].oper & MASK)
			{
				mark = okcdata::DISABLED;
			}
		} 

		if (mark == okcdata::REGULAR)
			color = interface_color[DATA];
	}
//	std::cout << "Done ResetLineColors" << std::endl;
}



BrushMgr::~BrushMgr(void)
{

}

// returns min{ i | A[i] >= v} where A is a sorted array
int BrushMgr::SearchLowerBound(int dim, double v) const
{
	ARR<double> & databuf = Dataset.data_buf;
    const ARR<int> & thisIndex = Index[dim];
	const int width = Dataset.all_dims;
	const int len = thisIndex.Len();

	if (len == 0)
	{
		// Special case: this dimension is degenerate since
		// all values are the same. So it's all or nothing.
		// Check an arbitrary line to see if all lines are
		// >= to v (which is ok because all lines have the
		// same value at this dimension).
		return v > Dataset.data_buf[0 + dim] ? -1 : 0;
	}

	// This nastiness will access the data for line X
	//  at the current dimension (dim).
#define _data(X) (databuf[thisIndex[(X)] * width + dim])

	if (v > _data(len-1))
		return -1;

	// This is a variant of the classic binary search algorithm.
	// The difference is that the algorithm does not immediately
	// stop if v is found, and not finding v in the array does
	// not result in an error.

	int i = 0, j = len-1, k;
	double d;

	while(i <= j)
	{
		k = (i + j) >> 1;
		d = _data(k);
		if (d == v)
		{
			// We have found the value in the array, but there could
			// be another member of the data that has the same value.

			// If we have reached the left end of the array or if 
			// there does not exist another value to the left of [k],
			// k must be the index of the first value >= to v.
			if (k == 0 || _data(k-1) != v)
				return k;
			else
				// There is another value [k-1] which equals v. In case
				// the data has a lot of values equal to v, just iterate
				// normally within a range that includes [k-1]. This
				// makes sure that the worst-case input is handled in
				// O(log(N)) rather than O(N) if we just scanned to the
				// left.
				j = k-1;
		}
		else if (v < d)
			j = k-1;
		else
			i = k+1;
	}
	return i;
#undef _data
}

// returns max{ i | A[i] <= v} where A is a sorted array
int BrushMgr::SearchUpperBound(int dim, double v) const
{
	ARR<double> & databuf = Dataset.data_buf;
    const ARR<int> & thisIndex = Index[dim];
	const int width = Dataset.all_dims;
	const int len = thisIndex.Len();

	if (len == 0)
	{
		// Special case: this dimension is degenerate since
		// all values are the same. So it's all or nothing.
		// Check an arbitrary line to see if all lines are
		// <= to v (which is ok because all lines have the
		// same value at this dimension).
		return v < Dataset.data_buf[0 + dim] ? -1 : 0;
	}

	// This nastiness will access the data for line i
	//  at the current dimension (dim).
#define _data(i) (databuf[thisIndex[i] * width + dim])

	if (v < _data(0))
		return -1;

	// This is a variant of the classic binary search algorithm.
	// The difference is that the algorithm does not immediately
	// stop if v is found, and not finding v in the array does
	// not result in an error.

	int i = 0, j = len-1, k;
	double d;
	while(i <= j)
	{
		k = (i + j + 1) >> 1;
		d = _data(k);
		if (d == v)
		{
			// We have found the value in the array, but there could
			// be another member of the data that has the same value.

			// If we have reached the right end of the array or if 
			// there does not exist another value to the right of [k],
			// k must be the index of the last value <= to v.
			if (k == len-1 || _data(k+1) != v)
				return k;
			else
				// There is another value [k+1] which equals v. In case
				// the data has a lot of values equal to v, just iterate
				// normally within a range that includes [k+1]. This
				// makes sure that the worst-case input is handled in
				// O(log(N)) rather than O(N) if we just scanned to the
				// right.
				i = k+1;
		}
		else if (v < d)
			j = k-1;
		else
			i = k+1;
	}
	return j;
#undef _data
}

