// CHierNavHelperMM.cpp: implementation of the CHierNavHelperMM class.
//
//////////////////////////////////////////////////////////////////////

#include<iostream>

#include <xmdv/XmdvTool.h>
#include <xmdv/CHierNavHelperMM.h>
#include <xmdv/CHierHist.h>

CHierNavHelperMM::CHierNavHelperMM()
{

}

CHierNavHelperMM::~CHierNavHelperMM()
{

}


/// 
/// A function that finds the number of levels in the cluster tree
/// 
/// @return height of the tree
/// 
int CHierNavHelperMM ::  FindTreeHeight()
{
	int heightOfTheTree = 0;

	this->FindTreeHeight(&g_globals.btree->nodes[0],0,&heightOfTheTree);

	return(heightOfTheTree);
}

/// 
/// A function that evaluates the level values and populates the array
/// contour radius
/// 
/// @param countour_radius
/// 
void CHierNavHelperMM :: FindRadiusDistribution(ARR<double> &contour_radius)
{
	assert(CHierHist::histo.Len()!=0);

    int MIN_GAP = 5;

    double R = g_globals.btree->nodes[0].sqrt_radius;

    contour_radius.SetLen(0);

    for (int i=CHierHist::histo.Len()-1, previ=1000000;i>=0;i--)
    {
        if (CHierHist::histo[i]==0) continue;
        if ((previ-i)<MIN_GAP) continue;

        contour_radius.InsertTail( R*i/double(CHierHist::histo.Len()-1) );
        previ = i;
    }
    contour_radius.InsertTail(0.0);
}

/// 
/// A function that finds the tree height. This is a recursive function 
/// that traverses the cluster_tree to the deepest level to find the 
/// height of the cluster tree.
/// 
void CHierNavHelperMM :: FindTreeContour(ARR< ARR<double> > &tree_contour,
										 ARR< ARR<double> > &tree_color,
										 ARR<double> contour_radius)
{	
    for (int i=0;i<tree_contour.Len();i++) {
        FindTreeContour(&g_globals.btree->nodes[0],
						0,
						contour_radius[i],
						tree_contour[i],
						tree_color[i],
						0,
						1);
	//Just for debugging remove this once finished debugging
	/*using namespace std;
	cout<<endl<<endl;	
	cout<<"********************************************"<<endl;	
	cout<<" The contour calculated by the front end for :"<<contour_radius[i]<<endl;
	debugArr(tree_contour[i]);
	cout<<"---------------------------------------------*"<<endl;
	cout<<endl<<endl;
	cout<<"The color calculated for the contour "<<"\n";
	debugArr(tree_color[i]);*/
	
    }
	
}

///
/// A function that returns the and evaluates the tree contour. The left 
/// and the right extent indicates what elements of the array are to be set
/// by the current call
/// 
/// @param node, a pointer to the root node
/// @param curr_ht, gives the current level of depth
/// @param rad, gives the level value that we are looking for using recursion
/// @param contour, is one dimensional array that stores the value of the evaluated contour
/// @param L, is the left extent
/// @param R extent Right extent
///  
void CHierNavHelperMM :: FindTreeContour(Cluster *node,
					 int curr_ht,
					 double rad,
					 ARR<double> &contour,
					 ARR<double> &color,
					 double L,
					 double R)
{
    if (node->parent != -1)
    {
        Cluster *parent = &g_globals.btree->nodes[node->parent];

        //
        // If it's a clustered node and a leaf
        //
        if (node->num_children==0 && node->sqrt_radius >= rad) {
            rad = node->sqrt_radius+ 0.01; // (1.0/(double)CONTOUR_RES);
        }

        if (node->sqrt_radius < rad)
        {
            //
            // Iso-radius contour cuts between node and parent.
            // "interp" is the intersection ratio between node and parent.
            // "y" is the intersection height of the contour.
            //

            double interp;

            if (parent->sqrt_radius - node->sqrt_radius < 0.0000001) interp = 0;
            else
                interp = (rad - node->sqrt_radius) /
                           (parent->sqrt_radius - node->sqrt_radius);

            double y = curr_ht - interp;

            //
            // Find the contour segment(s) this node covers: [x0,x1]
            //
            double x0 = L*CONTOUR_RES;
            double x1 = R*CONTOUR_RES;

            int xi0 = (int) ceil(x0+.000000001);
            int xi1 = (int) floor(x1-.00000001);

            for (int i=xi0;i<xi1;i++) {
                // Coverage that fills whole segments
                assert(contour[i] == 0.0);
                contour[i] = y;
                color[i] = (L+R)*.5;
            }

            if (xi0>0 && x1>xi0) {
                // Left end of the coverage
                double coverage = (xi0-x0);
                contour[xi0-1] += coverage*y;
                if (coverage>.9) color[xi0-1] = (L+R)*.5;
            }
            if (xi1<CONTOUR_RES && x0<xi1) {
                // Right end of the coverage
                double coverage = (x1-xi1);
                contour[xi1] += coverage*y;
                if (coverage>.9) color[xi1] = (L+R)*.5;
            }
            if (xi0>=x1 && xi1<=x0) {
                // Partial coverage
                double coverage = (x1-x0);
                contour[xi1] += coverage*y;
                if (coverage>.9) color[xi1] = (L+R)*.5;
            }

            return;
        }
    }

    double k = (R-L)/(double)(node->num_children);

    for (int i=0; i<node->num_children; i++) {
        FindTreeContour(&g_globals.btree->nodes[node->child[i]], 
						curr_ht+1,rad,contour, color, L+i*k, L+(i+1)*k);
    } // for (int i=0; i<node->num_children; i++)	
}

/// 
/// A function that finds the tree height
/// This is a recursive function that traverses the cluster_tree to
/// to the deepest level to find the height of the cluster tree.
/// 
void CHierNavHelperMM :: FindTreeHeight(Cluster* node, 
										int curr_ht, 
										int *max_ht)
{
    if (curr_ht>*max_ht) *max_ht = curr_ht;

    for (int i=0; i<node->num_children; i++)
        FindTreeHeight(&g_globals.btree->nodes[node->child[i]], curr_ht+1, max_ht);
}

/*
 *	Returns the radius of the root node.
 */
double CHierNavHelperMM::FindRootRadius()
{
	return(g_globals.btree->nodes[0].sqrt_radius);
}