/* file: expAverage.c 
 * purpose: define Exponential Weighted Average class
 * revision log:
 * algorithm: 
 *   Ave[1] = x1. Ave[t] = alpha*x[t] + (1-alpha)*Ave[t-1]
 *   where alpha = smoothing parameter between [0,1].
 *   example: x1=1,  Ave[1]=1,    alpha=0.5
 *            x2=2,  Ave[2]=0.5*2 + 0.5*1 = 1.5
 *            x3=3,  Ave[3]=0.5*3 + 0.5*1.5 = 0.5*3 + 0.5*0.5*2 + 0.5*0.5*1 = 2.25
 *            x4=4,  Ave[4]=0.5*4 + 0.5*2.25 = 3.125
 */

#include <expAverage.h>
#include <stdlib.h>
#include <stdio.h>

struct expAverage_Str
{	float alpha ;  /* smoothing factor */
	float prevAve ;   /* previous average for the queries seen so far */
	int initFlag ;   /* 1 if average already initialized */
} expAverage_Str ;


expAverage expAverage_Create()
{ 
	expAverage g ;

	g = (expAverage) malloc(sizeof(struct expAverage_Str)) ;
	if (g != NULL)
	{ 	g->alpha = 0.75 ;           // default smoothing factor
		g->prevAve = 0.0 ;
		g->initFlag = 0 ;
	}
	return(g) ;
}


void expAverage_Destroy(expAverage g) 
{ 
	free(g) ;
}


void expAverage_SetAlpha(expAverage g, float alpha)
{
	if (alpha >= 0.0 && alpha <= 1.0)
		g->alpha = alpha ;
	else
		printf("ExpAverage Error: alpha out of range.\n") ;
		 /* old alpha value retained */
}


float expAverage_Update(expAverage g, float m) 
{
	if (g->initFlag == 1)
	    g->prevAve = g->alpha*m + (1-g->alpha)*g->prevAve ;
	else 
	{
		g->prevAve = m ;
		g->initFlag = 1 ;
	}
	return(g->prevAve) ;
}


float expAverage_GetValue(expAverage g) 
{
	return(g->prevAve) ;
}


void expAverage_Test()
{
	expAverage g ;

	printf("Testing expAverage class -- begin \n") ;
	if ((g = expAverage_Create()) == NULL)
		printf("No more memory\n") ;
	else
	{
        expAverage_SetAlpha(g, 0.5) ;
		printf("+ 0.0, average=%f \n", expAverage_Update(g, 0.0)) ;
		printf("+ 1.0, average=%f \n", expAverage_Update(g, 1.0)) ;
		printf("+ 2.0, average=%f \n", expAverage_Update(g, 2.0)) ;
		printf("+ 3.0, average=%f \n", expAverage_Update(g, 3.0)) ;
		printf("+ 4.0, average=%f \n", expAverage_Update(g, 4.0)) ;
	}

	expAverage_Destroy(g) ;
	printf("Testing expAverage class -- done \n") ;
	return ;
}
